/*
 * Grammar for a simple interface description language.
 *
 * Copyright (C) 2019 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

%{
#include <stdio.h>
#include <string.h>
#include "program.h"
#include "parser.h"

/* Standard signatures. */

int yylex(void);
void yyerror(const char *);

/* State information, also reset by the function defined below. */

struct interface *last_interface = NULL;
struct include *last_include = NULL;
%}

%define parse.error verbose

/* Define the identifier type employed in the scanner. */

%union {
  long num;
  char *str;
  struct include inc;
  struct interface iface;
  struct signature sig;
  struct identifier ident;
  struct parameter param;
  struct attribute attr;
}

/* Associate identifiers and similar values with the string union member. */

%type <str> IDENTIFIER DECIMAL HEXADECIMAL
%type <str> interface_name
%type <str> qualifier operation
%type <str> HEADER
%type <str> attr_param

/* Specifiers and items have numeric values. */

%type <num> IN INOUT OUT
%type <num> specifier
%type <num> CAP FPAGE
%type <num> item

/* Associate rules with the appropriate union members. */

%type <inc> include
%type <iface> q_interface
%type <iface> interface
%type <sig> q_signature
%type <sig> signature
%type <sig> signatures
%type <attr> attribute
%type <attr> attributes
%type <param> parameter
%type <param> parameters
%type <ident> attr_params
%type <ident> identifiers

/* Scanner token types. */

%token INCLUDE HEADER
%token INTERFACE
%token BEGINSECTION ENDSECTION
%token LPAR RPAR
%token LBRACE RBRACE
%token SEP TERM
%token IN INOUT OUT
%token CAP FPAGE
%token IDENTIFIER DECIMAL HEXADECIMAL

/* Starting rule. */

%start file

%%

file            : statements
                  { write_files(last_interface); }
                ;

statements      : statement statements
                | %empty
                ;

statement       : include
                | q_interface
                  { $1.tail = last_interface; last_interface = copy_interface($1); }
                ;

include         : INCLUDE HEADER
                  { $$.filename = $2; $$.tail = last_include; last_include = copy_include($$); }
                ;

/* Interface qualified with or without attributes. */

q_interface     : LBRACE attributes RBRACE interface
                  { $4.attributes = copy_attribute($2); $$ = $4; }
                | interface
                  { $$.attributes = NULL; }
                ;

interface       : INTERFACE interface_name BEGINSECTION signatures ENDSECTION TERM
                  { $$.name = $2; $$.signatures = copy_signature($4); $$.includes = last_include; last_include = NULL; }
                | INTERFACE interface_name BEGINSECTION ENDSECTION TERM
                  { $$.name = $2; $$.signatures = NULL; $$.includes = last_include; last_include = NULL; }
                ;

interface_name  : IDENTIFIER ;

/* Signatures qualified with or without attributes. */

signatures      : q_signature signatures
                  { copy_tail(signature, $1, $2); $$ = $1; }
                | q_signature
                  { $1.tail = NULL; $$ = $1; }
                ;

q_signature     : LBRACE attributes RBRACE signature
                  { $4.attributes = copy_attribute($2); $$ = $4; }
                | signature
                  { $1.attributes = NULL; $$ = $1; }
                ;

signature       : qualifier operation LPAR RPAR TERM
                  { $$.qualifier = $1; $$.operation = $2; $$.parameters = NULL; }
                | qualifier operation LPAR parameters RPAR TERM
                  { $$.qualifier = $1; $$.operation = $2; $$.parameters = copy_parameter($4); }
                ;

qualifier       : IDENTIFIER ;

operation       : IDENTIFIER ;

/* Attributes with or without parameters. */

attributes      : attribute SEP attributes
                  { copy_tail(attribute, $1, $3); $$ = $1; }
                | attribute
                  { $1.tail = NULL; $$ = $1; }
                ;

attribute       : IDENTIFIER LPAR attr_params RPAR
                  { $$.attribute = $1; $$.identifiers = copy_identifier($3); }
                | IDENTIFIER
                  { $$.attribute = $1; $$.identifiers = NULL; }
                ;

attr_params     : attr_param SEP attr_params
                  { $$.identifier = $1; copy_tail(identifier, $$, $3); }
                | attr_param
                  { $$.identifier = $1; $$.tail = NULL; }
                ;

attr_param      : IDENTIFIER
                | DECIMAL
                | HEXADECIMAL
                ;

/* Signature parameters. */

parameters      : parameter SEP parameters
                  { copy_tail(parameter, $1, $3); $$ = $1; }
                | parameter
                  { $1.tail = NULL; $$ = $1; }
                ;

parameter       : specifier item identifiers
                  { $$.specifier = $1; $$.cls = $2; $$.identifiers = copy_identifier($3); }
                | specifier identifiers
                  { $$.specifier = $1; $$.cls = WORD_CLASS; $$.identifiers = copy_identifier($2); }
                ;

specifier       : IN | INOUT | OUT ;

item            : CAP | FPAGE ;

identifiers     : IDENTIFIER
                  { $$.identifier = $1; $$.tail = NULL; }
                | IDENTIFIER identifiers
                  { $$.identifier = $1; copy_tail(identifier, $$, $2); }
                ;

%%

void *copy(void *value)
{
  union YYSTYPE *v;

  if (value == NULL)
    return value;

  v = (union YYSTYPE *) malloc(sizeof(union YYSTYPE));
  memcpy(v, value, sizeof(union YYSTYPE));
  return v;
}

void reset(void)
{
  last_interface = NULL;
  last_include = NULL;
}

void yyerror(const char *message)
{
  printf("%s\n", message);
}
